module Santa where

import Control.Concurrent.STM
import Control.Concurrent
import System.Random
import qualified MyBuffer as Buf

{-
  We need a buffer, because otherwise IO will be scrambled.
-}

type Buf = Buf.Buf String
bget = atomically . Buf.get
bput b s = atomically $ Buf.put b s
bnew = atomically $ Buf.new


{- Actions of elves and deer -}
meetInStudy :: Buf -> Int -> IO ()
meetInStudy buf id = bput buf $ 
  "Elf "++show id++" meeting in the study"

deliverToys :: Buf -> Int -> IO ()
deliverToys buf id = bput buf $ 
  "Reindeer "++show id++" delivering toys"

{- Both elves and deer perform a simple job:
    - join their respective group
    - pass through the entrance gate
    - perform the above task
    - exit through the second gate
-}
helper1 :: Group -> IO () -> IO ()
helper1 grp task = do 
  (inGate, outGate) <- joinGroup grp
  passGate inGate
  task
  passGate outGate
                      
elf1, reindeer1 :: Buf -> Group -> Int -> IO ()
elf1 buf grp elfId = 
  helper1 grp (meetInStudy buf elfId)
reindeer1 buf grp reinId = 
  helper1 grp (deliverToys buf reinId)

{- A gate has a total capacity and a current capacity, which can
   be concurrently modified. 
-}
data Gate = Gate Int (TVar Int)

{- Create a new gate with capacity n, but initical capacity 0.
   This is done so that Santa can explicitly open a gate, by 
   re-setting a gate's capacity (to full).
-}
newGate :: Int -> STM Gate
newGate n = do tv <- newTVar 0
               return $ Gate n tv
               
{- To pass through a gate, it must have capacity left. We decrease
   the capacity if we manage to pass through.
-}
passGate :: Gate -> IO ()
passGate (Gate n tv) = 
  atomically $ do c <- readTVar tv
                  check (c > 0)
                  writeTVar tv (c - 1)
  
{- Santa operates gates: First, he opens it, then he waits until
   all deer/elves have passed through.
   Here, we rely on elves/deer to pass through whenever possible
-}
operateGate :: Gate -> IO ()
operateGate (Gate n tv) = do
  atomically $ writeTVar tv n
  atomically $ do c <- readTVar tv
                  check (c == 0)

{- Elves/deer gather in a group, waiting for santa to open up the gates,
   which he does as soon as the group is large enough.
   Groups have a capacity (total and current) and two gates associated.
-}
data Group = Group Int (TVar (Int, Gate, Gate))

{- Creating a new group means creating two new gates and a group TVar.
-}
newGroup :: Int -> IO Group
newGroup n = atomically $ do
  g1 <- newGate n
  g2 <- newGate n
  tv <- newTVar (n, g1, g2)
  return $ Group n tv

{- A group can be joined as long as there's space left.
-}
joinGroup :: Group -> IO (Gate, Gate)
joinGroup (Group n tv) = 
  atomically $ do (k, g1, g2) <- readTVar tv
                  check (k > 0)
                  writeTVar tv (k - 1, g1, g2)
                  return $ (g1, g2)

{- (part of) Santa's job: wait for a group to be complete.
 Set up new gates for the group, so that other elves/deer can gather
 and wait for the next round.
 -}
awaitGroup :: Group -> STM (Gate, Gate)
awaitGroup (Group n tv) = do 
  (k, g1, g2) <- readTVar tv
  check (k == 0)
  g1' <- newGate n
  g2' <- newGate n
  writeTVar tv (n, g1', g2')
  return (g1, g2)

-- An elf does his elf thing, infinitely.
elf :: Buf -> Group -> Int -> IO ThreadId
elf buf grp id = forkIO $ forever $ do elf1 buf grp id
                                       randomDelay

-- So does a deer.
reindeer :: Buf -> Group -> Int -> IO ThreadId
reindeer buf grp id = forkIO $ forever $ do reindeer1 buf grp id
                                            randomDelay

{- Santa is the most interesting part:
     - he chooses a group ready for action, giving preference to his reindeer
     - operates the corresponding entrance gate
     - lets them do their work
     - operates the exit gate
 -}
santa :: Buf -> Group -> Group -> IO ()
santa buf elves deer = do
  (name, (g1, g2)) <- atomically $ chooseGroup "reindeer" deer `orElse` chooseGroup "elves" elves
  bput buf $ "Ho, ho, my dear " ++ name
  operateGate g1                            
  operateGate g2
  where chooseGroup :: String -> Group -> STM (String, (Gate, Gate))
        chooseGroup msg grp = do 
          gs <- awaitGroup grp
          return (msg, gs)
  
{-
  Let's run it: Start 10 elves that can join up in a group of 3,
  9 reindeer which must all be present in the group,
  and let santa do his job.
-}
main :: IO ()
main = do buf <- setupBufferListener 
          
          elfGroup <- newGroup 3
          sequence_ [ elf buf elfGroup id | id <- [1 .. 10] ]
          
          deerGroup <- newGroup 9
          sequence_ [ reindeer buf deerGroup id | id <- [1 .. 9]]
          
          forever (santa buf elfGroup deerGroup)

---------------------------------
-- Helper functions
---------------------------------

setupBufferListener :: IO Buf
setupBufferListener = do 
  b <- bnew
  forkIO $ forever $ do { ln <- bget b; putStrLn ln } 
  return b
  
forever :: IO () -> IO ()
forever m = do m
               forever m

randomDelay :: IO ()
randomDelay = do time <- getStdRandom (randomR (1, 1000000))
                 threadDelay time
                 