module Santa where

import Control.Concurrent.STM
import Control.Concurrent
import System.Random
import qualified MyBuffer as Buf

{-
  We need a buffer, because otherwise IO will be scrambled.
-}

type Buf = Buf.Buf String
bget = atomically . Buf.get
bput b s = atomically $ Buf.put b s
bnew = atomically $ Buf.new


{- Actions of elves and deer -}
meetInStudy :: Buf -> Int -> IO ()
meetInStudy buf id = bput buf $ 
  "Elf "++show id++" meeting in the study"

deliverToys :: Buf -> Int -> IO ()
deliverToys buf id = bput buf $ 
  "Reindeer "++show id++" delivering toys"

{- Both elves and deer perform a simple job:
    - join their respective group
    - pass through the entrance gate
    - perform the above task
    - exit through the second gate
-}
helper1 :: Group -> IO () -> IO ()
helper1 grp task = do 
  (inGate, outGate) <- joinGroup grp
  passGate inGate
  task
  passGate outGate
                      
elf1, reindeer1 :: Buf -> Group -> Int -> IO ()
elf1 buf grp elfId = 
  helper1 grp (meetInStudy buf elfId)
reindeer1 buf grp reinId = 
  helper1 grp (deliverToys buf reinId)

{- A gate has a total capacity and a current capacity, which can
   be concurrently modified. 
-}
data Gate = Gate Int (TVar Int)

newGate :: Int -> STM Gate
newGate n = do tv <- newTVar 0
               return $ Gate n tv
          
passGate :: Gate -> IO ()
passGate (Gate n tv) = 
  atomically $ do c <- readTVar tv
                  check (c > 0)
                  writeTVar tv (c - 1)
  
operateGate :: Gate -> IO ()
operateGate (Gate n tv) = do
  atomically $ writeTVar tv n
  atomically $ do c <- readTVar tv
                  check (c == 0)

data Group = Group Int (TVar (Int, Gate, Gate))

newGroup :: Int -> IO Group
newGroup n = atomically $ do
  g1 <- newGate n
  g2 <- newGate n
  tv <- newTVar (n, g1, g2)
  return $ Group n tv

joinGroup :: Group -> IO (Gate, Gate)
joinGroup (Group n tv) = 
  atomically $ do (k, g1, g2) <- readTVar tv
                  check (k > 0)
                  writeTVar tv (k - 1, g1, g2)
                  return $ (g1, g2)

awaitGroup :: Group -> STM (Gate, Gate)
awaitGroup (Group n tv) = do 
  (k, g1, g2) <- readTVar tv
  check (k == 0)
  g1' <- newGate n
  g2' <- newGate n
  writeTVar tv (n, g1', g2')
  return (g1, g2)

-- An elf does his elf thing, indefinitely.
elf :: Buf -> Group -> Int -> IO ThreadId
elf buf grp id = forkIO $ forever $ 
  do elf1 buf grp id
     randomDelay

-- So does a deer.
reindeer :: Buf -> Group -> Int -> IO ThreadId
reindeer buf grp id = forkIO $ forever $ 
  do reindeer1 buf grp id
     randomDelay

santa :: Buf -> Group -> Group -> IO ()
santa buf elves deer = do
  (name, (g1, g2)) <- atomically $ 
    chooseGroup "reindeer" deer `orElse` 
      chooseGroup "elves" elves
  bput buf $ "Ho, ho, my dear " ++ name
  operateGate g1                            
  operateGate g2

chooseGroup :: String -> Group -> 
               STM (String, (Gate, Gate))
chooseGroup msg grp = do 
  gs <- awaitGroup grp
  return (msg, gs)
  
main :: IO ()
main = do buf <- setupBufferListener 
          
          elfGroup <- newGroup 3
          sequence_ [ elf buf elfGroup id | 
                      id <- [1 .. 10] ]
          deerGroup <- newGroup 9
          sequence_ [ reindeer buf deerGroup id | 
                      id <- [1 .. 9]]
          forever (santa buf elfGroup deerGroup)

---------------------------------
-- Helper functions
---------------------------------

setupBufferListener :: IO Buf
setupBufferListener = do 
  b <- bnew
  forkIO $ forever $ do { ln <- bget b; putStrLn ln } 
  return b
  
forever :: IO () -> IO ()
forever m = do m
               forever m

randomDelay :: IO ()
randomDelay = do time <- getStdRandom (randomR (1, 1000000))
                 threadDelay time
                 